/*====================================================================
| Version: December 7, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package filters;

import ij.process.FloatProcessor;

public class CTLoG{

	private final double RADIUS_SIGMA_RATIO = 1.0/3.0; 
	private final double GAUSSIAN_TRUNCATION = 5.0;
	private int width, height;
	private float[] pixels1 = null;
	private float[] pixels2 = null;
	
	public CTLoG(){
	}
	
	/**
	* Apply a Laplacian of Gaussian 2D.
	* Separable implementation.
	*/
	public void runFilter(FloatProcessor ip, double radiusX, double radiusY) {
		double sigmaX = RADIUS_SIGMA_RATIO * radiusX;
		double sigmaY = RADIUS_SIGMA_RATIO * radiusY;

		FloatProcessor ip2 = (FloatProcessor)ip.duplicate();

		width = ip.getWidth();
		height = ip.getHeight();
		
		pixels1 = (float[])ip.getPixels();
		pixels2 = (float[])ip2.getPixels();
		
		double pd = 2.0*Math.PI;
		double sx = (sigmaX > 0 ? sigmaX: 1.0);
		double sy = (sigmaY > 0 ? sigmaY: 1.0);
		double cst = 1.0/(pd*sx*sy);

		double kernelFactX[] = createKernelLoG_Fact(sigmaX, cst);	
		double kernelBaseX[] = createKernelLoG_Base(sigmaX);	
		double kernelFactY[] = createKernelLoG_Fact(sigmaY, cst);	
		double kernelBaseY[] = createKernelLoG_Base(sigmaY);	

		if (sigmaY > 0.0) {
	 		double vinY[] = new double[height];
	 		double voutY[] = new double[height];
			for (int x=0; x<width; x++) {
				getY(pixels1, x, vinY);
				convolve(vinY, voutY, kernelFactY);
				putY(pixels1, x, voutY);
				
				getY(pixels2, x, vinY);
				convolve(vinY, voutY, kernelBaseY);
				putY(pixels2, x, voutY);
			}
		}
 		if (sigmaX > 0.0) {
	 		double vinX[] = new double[width];
	 		double voutX[] = new double[width];
			for (int y=0; y<height; y++) {
				getX(pixels1, y, vinX);
				convolve(vinX, voutX, kernelBaseX);
				putX(pixels1, y, voutX);
				
				getX(pixels2, y, vinX);
				convolve(vinX, voutX, kernelFactX);
				putX(pixels2, y, voutX);
			}
		}

 		int size = width*height;
 		for(int i=0; i<size; i++){
 			pixels1[i] += pixels2[i];
 		}
	
	}

	/**
	* cst*(x^2/(sigma^4)-1/(sigma^2))*exp(-(x^2)/(sigma^2))
	*/
	private double[] createKernelLoG_Fact(double sigma, double cst) {
		
		if (sigma <= 0.0) {
			double[] kernel = new double[1];
			kernel[0] = 1.0;
			return kernel;
		}
		
		double s2 = sigma*sigma;
		double s4 = s2*s2;
		double dem = 2.0*s2;
		int size = (int)Math.round(((int)(sigma*GAUSSIAN_TRUNCATION))*2.0 + 1);	// always odd size
		int size2 = size/2;
		double[] kernel = new double[size];
		
		double x;
		for(int k=0; k<size; k++) {
			x = (k-size2)*(k-size2);
			kernel[k] = -cst * (x/s4-1.0/s2) * Math.exp(-x/dem);
		}
		return kernel;
	}

	/**
	* exp(-(x^2)/(sigma^2))
	*/
	private double[] createKernelLoG_Base(double sigma) {
	
		if (sigma <= 0.0) {
			double[] kernel = new double[1];
			kernel[0] = 1.0;
			return kernel;
		}
		
		double s2 = sigma*sigma;
		double dem = 2.0*s2;
		int size = (int)Math.round(((int)(sigma*GAUSSIAN_TRUNCATION))*2.0 + 1);	// always odd size
		int size2 = size/2;
		double[] kernel = new double[size];
		
		double x;
		for(int k=0; k<size; k++) {
			x = (k-size2)*(k-size2);
			kernel[k] = Math.exp(-x/dem);
		}
		return kernel;
	}

	private void putX(float[] pixels, int y, double[] buffer) {
		int offset = y*width;
		int leni = buffer.length;
		for (int i=0; i<leni; i++) {
			pixels[offset] = (float)(buffer[i]); 
			offset++; 
		}
	}
	
	private void getX(float[] pixels, int y, double[] buffer) {
		int offset = y*width;
		int leni = buffer.length;
		for (int i=0; i<leni; i++) {
			buffer[i] = (double)(pixels[offset]);
			offset++;
		}
	}

	private void putY(float[] pixels, int x, double[] buffer) {
		int offset = x;
		int leni = buffer.length;
		for (int i=0; i<leni; i++) {
			pixels[offset] = (float)(buffer[i]);
			offset+=width;
		}
	}

	
	private void getY(float[] pixels, int x, double[] buffer) {
		int offset = x;
		int leni = buffer.length;
		for (int i=0; i<leni; i++) {
			buffer[i] = (double)(pixels[offset]);
			offset+=width;
		}
	}
	
	private double[] convolve(double vin[], double vout[], double kernel[]) {
		int n = vin.length;
		int nk = kernel.length;
		int kc = nk/2;
		int	period = (n <= 1 ? 1: 2*n - 2);
		
		int im;
		for (int i=0; i<n; i++) {
			double sum=0.0;
			for (int k=0; k<nk; k++) {
				im = i + k - kc;
				while (im < 0)
					im += period;
				while (im >= n) {
					im = period - im;
					im = (im < 0 ? -im : im);
				}
				sum += kernel[k] * vin[im];
			}
			vout[i] = sum;
		}
		return vout;
	}
 }


