/*====================================================================
| Version: November 11, 2010
\===================================================================*/

/*====================================================================
| Ricard Delgado-Gonzalo
| EPFL/STI/IOA/LIB/BM.4.141
| e-mail: ricard.delgado@gmail.com
| URL: http://www.ricarddelgado.com/
\===================================================================*/
package segmentation;

import javax.swing.JTextArea;

import ij.IJ;
import ij.process.ByteProcessor;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;

public class CTWatershedProcessing {

	private static int MAXBASINS = 200000;
	private int off;
	private boolean animation 	= false;
	public static float MINLABEL = 10.0f;
	private JTextArea taskLog = null;
	
	private boolean[] regionMask = null;
	private double[] centroidsX = null;
	private double[] centroidsY = null;
	private int[] xmin = null;
	private int[] ymin = null;
	private int[] xmax = null;
	private int[] ymax = null;
	private double[] area = null;
	private boolean[] markedRegions = null;
	private boolean[][] neighborhoodMatrix = null;
	private int Nregions;
	
	private FloatProcessor output = null;
	private ByteProcessor image = null;
	private int width;
	private int height;
	
	public CTWatershedProcessing(ImageProcessor ip, boolean[] regionMask, int minVal, int maxVal, int regionOffset) {
		off = regionOffset;
		float[] data = (float[])ip.getPixels();
		int[] dataf = new int[data.length];
		
		for(int i=0; i<data.length; i++){
			dataf[i] = (int) (255.0*(data[i]-minVal)/((float)(maxVal-minVal)));
		}
		
		FloatProcessor newFloat = new FloatProcessor(ip.getWidth(), ip.getHeight(), dataf);
		newFloat.resetMinAndMax();
		image = (ByteProcessor)newFloat.convertToByte(false);
		width = image.getWidth();
		height = image.getHeight();
		this.regionMask = regionMask;
	}	

	public void setAnimation(boolean animation) {
		this.animation = animation;
	}

	public void setLog(JTextArea taskLog){
		this.taskLog = taskLog;
	}

	private void writeLog(String str){
		if(taskLog!=null){
			taskLog.append(str+"\n");
			taskLog.setCaretPosition(taskLog.getDocument().getLength());
		}else{
			IJ.log(str);
		}
	}
	
	public boolean watershed() {
		ByteProcessor inputDuplicate = (ByteProcessor)image.duplicate();
		CTWatershed ws = new CTWatershed();
		ws.setLog(taskLog);
		if (animation) {
			ws.enableAnimation();
		}
		output = ws.doWatershed(inputDuplicate);
		findNregions();
		writeLog("Detecting marked regions...");
		findMarkedRegions();
		writeLog("Deleting marked regions...");
		deleteMarkedRegions();
		if(Nregions>0) {
			findNeighborhood();
			measureBasins();
			return true;
		}
		return false;
	}

	private void deleteMarkedRegions(){
		int[] mapVector = new int[Nregions];
		int label = 1;
		for(int i=0; i<Nregions; i++){
			if(markedRegions[i]){
				mapVector[i] = -1;
			}else{
				mapVector[i] = label;
				label++;
			}
		}
		
		int l;
		int width1 = width-1;
		int height1 = height-1;
		
		for (int x=1; x<width1; x++){
			for (int y=1; y<height1; y++) {
				l = (int)Math.round(output.getPixelValue(x, y));
				if(l!=0){
					l -= (MINLABEL+1);
					if(mapVector[l]==-1){
						output.putPixelValue(x, y, 0.0);
					}else{
						output.putPixelValue(x,y,mapVector[l]+MINLABEL);
					}
					
					
					
				}
			}
		}
		findNregions();
	}
	
	private void findNregions(){
		output.resetMinAndMax();
		int maxClass = (int)output.getMax();
		Nregions = (int)Math.round(maxClass-MINLABEL);
		if(Nregions<=0){
			writeLog("No regions left");
		}
	}
	
	public boolean[][] getNeighborhoodMatrix(){
		return neighborhoodMatrix;
	} 
	
	private void findNeighborhood(){
	
		neighborhoodMatrix = new boolean[Nregions][Nregions];

		for(int i=0; i<Nregions; i++){
			for(int j=0; j<Nregions; j++){
				neighborhoodMatrix[i][j] = false;
			}
		}
		
		int l;
		int width2 = width-2;
		int height2 = height-2;
		for (int x=2; x<width2; x++){
			for (int y=2; y<height2; y++) {
				l = (int)Math.round(output.getPixelValue(x, y));
				if(l!=0){
					l -= (MINLABEL+1);
					int l1 = (int)Math.round(output.getPixelValue(x+2, y-2));
					int l2 = (int)Math.round(output.getPixelValue(x+2, y-1));
					int l3 = (int)Math.round(output.getPixelValue(x+2, y));
					int l4 = (int)Math.round(output.getPixelValue(x+2, y+1));
					int l5 = (int)Math.round(output.getPixelValue(x+2, y+2));
					
					
					int l6 = (int)Math.round(output.getPixelValue(x-2, y-2));
					int l7 = (int)Math.round(output.getPixelValue(x-2, y-1));
					int l8 = (int)Math.round(output.getPixelValue(x-2, y));
					int l9 = (int)Math.round(output.getPixelValue(x-2, y+1));
					int l10 = (int)Math.round(output.getPixelValue(x-2, y+2));
				
					
					int l11 = (int)Math.round(output.getPixelValue(x-1, y-2));
					int l12 = (int)Math.round(output.getPixelValue(x, y-2));
					int l13 = (int)Math.round(output.getPixelValue(x+1, y-2));
					
					int l14 = (int)Math.round(output.getPixelValue(x-1, y+2));
					int l15 = (int)Math.round(output.getPixelValue(x, y+2));
					int l16 = (int)Math.round(output.getPixelValue(x+1, y+2));			
					
					
					if(l1!=0) {
						l1 -= (MINLABEL+1);
						neighborhoodMatrix[l][l1] = true;
					}

					if(l2!=0) {
						l2 -= (MINLABEL+1);
						neighborhoodMatrix[l][l2] = true;
					}
					
					if(l3!=0) {
						l3 -= (MINLABEL+1);
						neighborhoodMatrix[l][l3] = true;
					}
					
					if(l4!=0) {
						l4 -= (MINLABEL+1);
						neighborhoodMatrix[l][l4] = true;
					}
					
					if(l5!=0) {
						l5 -= (MINLABEL+1);
						neighborhoodMatrix[l][l5] = true;
					}
					
					if(l6!=0) {
						l6 -= (MINLABEL+1);
						neighborhoodMatrix[l][l6] = true;
					}
					
					if(l7!=0) {
						l7 -= (MINLABEL+1);
						neighborhoodMatrix[l][l7] = true;
					}
					
					if(l8!=0) {
						l8 -= (MINLABEL+1);
						neighborhoodMatrix[l][l8] = true;
					}
					
					if(l9!=0) {
						l9 -= (MINLABEL+1);
						neighborhoodMatrix[l][l9] = true;
					}
					
					if(l10!=0) {
						l10 -= (MINLABEL+1);
						neighborhoodMatrix[l][l10] = true;
					}
					
					if(l11!=0) {
						l11 -= (MINLABEL+1);
						neighborhoodMatrix[l][l11] = true;
					}
					
					if(l12!=0) {
						l12 -= (MINLABEL+1);
						neighborhoodMatrix[l][l12] = true;
					}
					
					if(l13!=0) {
						l13 -= (MINLABEL+1);
						neighborhoodMatrix[l][l13] = true;
					}
					
					if(l14!=0) {
						l14 -= (MINLABEL+1);
						neighborhoodMatrix[l][l14] = true;
					}
					
					if(l15!=0) {
						l15 -= (MINLABEL+1);
						neighborhoodMatrix[l][l15] = true;
					}
					
					if(l16!=0) {
						l16 -= (MINLABEL+1);
						neighborhoodMatrix[l][l16] = true;
					}
				}
			}
		}
	}
	
	
	private void findMarkedRegions(){
		markedRegions = new boolean[Nregions];
		int l;
		int width1 = width-1;
		int height1 = height-1;
		for (int x=1; x<width1; x++){
			for (int y=1; y<height1; y++) {
				l = (int)Math.round(output.getPixelValue(x, y));
				if(l!=0){
					if(regionMask[x+ width*y]){
						l -= (MINLABEL+1);
						markedRegions[l] = true;	
					}
				}
			}
		}
	}
	
	public FloatProcessor getLabelsMap(){
		return output;
	}
	
	public void displayDams(){
		CTWatershedDisplay.showDams(output);
	}

	public void displayBasins(){
		CTWatershedDisplay.showBasins(output, CTWatershedDisplay.createLUTColor(MAXBASINS));
	}
	
	public void displayLabelizedBasins(){
		CTWatershedDisplay.showLabelizedBasins(output);
	}
	
	public void displayCentroids(){
		CTWatershedDisplay.showGravityCenters(output, centroidsX, centroidsY);
	}

	public void measureBasins() {
		area = new double[Nregions];
		centroidsX = new double[Nregions];
		centroidsY = new double[Nregions];
		xmin = new int[Nregions];
		ymin = new int[Nregions];
		xmax = new int[Nregions];
		ymax = new int[Nregions];
		
		
		int width1 = width-1;
		int height1 = height-1;
		for (int l=0; l<Nregions; l++) {
			area[l] = 0;
			centroidsX[l] = 0.0;
			centroidsY[l] = 0.0;
		
			xmin[l] = width1;
			ymin[l] = height1;
		
			xmax[l] = 0;
			ymax[l] = 0;

			
		}
		int l;
		for (int x=1; x<width1; x++){
			for (int y=1; y<height1; y++) {
				l = (int)Math.round(output.getPixelValue(x, y));
				if(l!=0){
					l -= (MINLABEL+1);
					if(x<xmin[l]) xmin[l] = x;
					if(y<ymin[l]) ymin[l] = y;
					if(x>xmax[l]) xmax[l] = x;
					if(y>ymax[l]) ymax[l] = y;
					centroidsX[l] += x;
					centroidsY[l] += y;
					area[l]++;
				}
			}
		}
		
		for (l=0; l<Nregions; l++) {
			if (area[l] > 0) {
				centroidsX[l]  /= (double)area[l];
				centroidsY[l] /= (double)area[l];
			}
		}
		
		
		for(int i=0; i<Nregions; i++){
			xmin[i] -= off; 
			if(xmin[i]<0) xmin[i] = 0;
			
			ymin[i] -= off;
			if(ymin[i]<0) ymin[i] = 0;

			xmax[i] += off; 
			if(xmax[i]>width) xmax[i] = width1;

			ymax[i] += off;
			if(ymax[i]>height) ymax[i] = height1;
		}
		
		
	}

	public int[] getXminRegions(){
		return xmin;
	}

	public int[] getYminRegions(){
		return ymin;
	}

	public int[] getXmaxRegions(){
		return xmax;
	}

	public int[] getYmaxRegions(){
		return ymax;
	}
	
	public int getNregions(){
		return Nregions;
	}
	
	public double[] getArea(){
		return area;
	}
	
	public double[] getCentroidsX(){
		return centroidsX;
	}
	
	public double[] getCentroidsY(){
		return centroidsY;
	}
	
}	